@@ -23,7 +23,7 @@ my $beginComment = '%BeginExifToolUpdate';
 my $endComment   = '%EndExifToolUpdate ';
 
 my $keyExt;     # crypt key extension
-my $pdfVer;     # version of PDF file we are currently writing
+my $pdfVer;     # version of PDF file we are writing (highest Version in Root dictionaries)
 
 # internal tags used in dictionary objects
 my %myDictTags = (
@@ -81,6 +81,7 @@ sub WritePDFValue($$$)
         EncodeString(\$val);
     } elsif ($format eq 'date') {
         # convert date to "D:YYYYmmddHHMMSS+-HH'MM'" format
+        $val =~ s/(:\d{2})\.\d*/$1/;              # remove sub-seconds
         $val =~ s/([-+]\d{2}):(\d{2})/${1}'${2}'/;  # change timezone delimiters if necessary
         $val =~ tr/ ://d;                       # remove spaces and colons
         $val =  "D:$val";                       # add leading "D:"
@@ -184,11 +185,19 @@ sub GetFreeEntries($)
 {
     my $dict = shift;
     my %xrefFree;
-    # from the start we have only written xref stream entries in 'CNn' format,
-    # so we can simplify things for now and only support this type of entry
+    # we write xref stream entries in 'CNn' or 'CNNn' format (with 8-byte 'NN' offset),
     my $w = $$dict{W};
-    if (ref $w eq 'ARRAY' and "@$w" eq '1 4 2') {
-        my $size = $$dict{_entry_size}; # this will be 7 for 'CNn'
+    if (ref $w eq 'ARRAY') {
+        my $bytes = "@$w";
+        my $fmt;
+        if ($bytes eq '1 4 2') {
+            $fmt = 'CNn';
+        } elsif ($bytes eq '1 8 2') {
+            $fmt = 'CNNn';
+        } else {
+            return \%xrefFree;
+        }
+        my $size = $$dict{_entry_size}; # this will be 7 for 'CNn' or 11 for 'CNNn'
         my $index = $$dict{Index};
         my $len = length $$dict{_stream};
         # scan the table for free objects
@@ -200,7 +209,12 @@ sub GetFreeEntries($)
             my $count = $$index[$i*2+1];
             for ($j=0; $j<$count; ++$j) {
                 last if $pos + $size > $len;
-                my @t = unpack("x$pos CNn", $$dict{_stream});
+                my @t = unpack("x$pos $fmt", $$dict{_stream});
+                if (@t == 4) {
+                    $t[1] = $t[1] * 4294967296 + $t[2];
+                    $t[2] = $t[3];
+                    @t = 3;
+                }
                 # add entry if object was free
                 $xrefFree{$start+$j} = [ $t[1], $t[2], 'f' ] if $t[0] == 0;
                 $pos += $size;  # step to next entry
@@ -290,23 +304,18 @@ sub WritePDF($$)
     $raf->Seek($pos, 0);
 
     # create a new ExifTool object and use it to read PDF and XMP information
-    my $newTool = new Image::ExifTool;
+    my $newTool = Image::ExifTool->new;
     $newTool->Options(List => 1);
     $newTool->Options(Password => $et->Options('Password'));
+    $newTool->Options(NoPDFList => $et->Options('NoPDFList'));
     $$newTool{PDF_CAPTURE} = \%capture;
     my $info = $newTool->ImageInfo($raf, 'XMP', 'PDF:*', 'Error', 'Warning');
     # not a valid PDF file unless we got a version number
-    # (note: can't just check $$info{PDFVersion} due to possibility of XMP-pdf:PDFVersion)
-    my $vers = $newTool->GetInfo('PDF:PDFVersion');
-    ($pdfVer) = values %$vers;
+    $pdfVer = $$newTool{PDFVersion};
     $pdfVer or $et->Error('Missing PDF:PDFVersion'), return 0;
     # check version number
-    if ($pdfVer > 1.7) {
-        if ($pdfVer >= 2.0) {
-            $et->Error("Can't yet write PDF version $pdfVer"); # (future major version changes)
-            return 1;
-        }
-        $et->Warn("ExifTool is untested with PDF version $pdfVer files", 1);
+    if ($pdfVer > 2.0) {
+        $et->Error("Writing PDF $pdfVer is untested", 1) and return 0;
     }
     # fail if we had any serious errors while extracting information
     if ($capture{Error} or $$info{Error}) {
@@ -396,6 +405,10 @@ sub WritePDF($$)
 
     # must pre-determine Info reference to be used in encryption
     my $infoRef = $prevInfoRef || \ "$nextObject 0 R";
+    unless (ref $infoRef eq 'SCALAR') {
+        $et->Error("Info dictionary is not an indirect object");
+        return $rtn;
+    }
     $keyExt = $$infoRef;
 
     # must encrypt all values in dictionary if they came from an encrypted stream
@@ -409,6 +422,9 @@ sub WritePDF($$)
     my $tagID;
     foreach $tagID (sort keys %$newTags) {
         my $tagInfo = $$newTags{$tagID};
+        if ($pdfVer >= 2.0 and not $$tagInfo{PDF2}) {
+            next if $et->Warn("Writing PDF:$$tagInfo{Name} is deprecated for PDF 2.0 documents",2);
+        }
         my $nvHash = $et->GetNewValueHash($tagInfo);
         my (@vals, $deleted);
         my $tag = $$tagInfo{Name};
@@ -531,7 +547,7 @@ sub WritePDF($$)
     }
     if ($metaChanged) {
         if ($newXMP) {
-            unless ($metaRef) {
+            unless (ref $metaRef) {
                 # allocate new PDF object
                 $metaRef = \ "$nextObject 0 R";
                 ++$nextObject;
@@ -655,24 +671,37 @@ sub WritePDF($$)
             $newXRef{$nextObject++} = [ Tell($outfile) - $$et{PDFBase} + length($/), 0, 'n' ];
             $$mainDict{Size} = $nextObject;
             # create xref stream and Index entry
-            $$mainDict{W} = [ 1, 4, 2 ];    # int8u, int32u, int16u ('CNn')
-            $$mainDict{Index} = [ ];
-            $$mainDict{_stream} = '';
-            my @ids = sort { $a <=> $b } keys %newXRef;
-            while (@ids) {
-                my $startID = $ids[0];
-                for (;;) {
-                    $id = shift @ids;
-                    my ($pos, $gen, $type) = @{$newXRef{$id}};
-                    if ($pos > 0xffffffff) {
-                        $et->Error('Huge files not yet supported');
-                        last;
+            my $bits = 4;
+Restart:    for (;;) {
+                $$mainDict{W} = [ 1, $bits, 2 ];    # int8u, int32u/int64u, int16u ('CNn' or 'CNNn')
+                $$mainDict{Index} = [ ];
+                $$mainDict{_stream} = '';
+                my @ids = sort { $a <=> $b } keys %newXRef;
+                while (@ids) {
+                    my $startID = $ids[0];
+                    for (;;) {
+                        $id = shift @ids;
+                        my ($pos, $gen, $type) = @{$newXRef{$id}};
+                        if ($pos > 0xffffffff) {
+                            if ($bits == 4) {
+                                # switch to 64-bit integer offsets
+                                $bits = 8;
+                                next Restart;
+                            }
+                        }
+                        if ($bits == 4) {
+                            $$mainDict{_stream} .= pack('CNn', $type eq 'f' ? 0 : 1, $pos, $gen);
+                        } else {
+                            my $hi = int($pos / 4294967296);
+                            my $lo = $pos - $hi * 4294967296;
+                            $$mainDict{_stream} .= pack('CNNn', $type eq 'f' ? 0 : 1, $hi, $lo, $gen);
+                        }
+                        last if not @ids or $ids[0] != $id + 1;
                     }
-                    $$mainDict{_stream} .= pack('CNn', $type eq 'f' ? 0 : 1, $pos, $gen);
-                    last if not @ids or $ids[0] != $id + 1;
+                    # add Index entries for this section of the xref stream
+                    push @{$$mainDict{Index}}, $startID, $id - $startID + 1;
                 }
-                # add Index entries for this section of the xref stream
-                push @{$$mainDict{Index}}, $startID, $id - $startID + 1;
+                last;
             }
             # write the xref stream object
             $keyExt = "$id 0 obj";  # (set anyway, but xref stream should NOT be encrypted)
@@ -751,7 +780,7 @@ C<PDF-update> pseudo group).
 
 =head1 AUTHOR
 
-Copyright 2003-2018, Phil Harvey (phil at owl.phy.queensu.ca)
+Copyright 2003-2025, Phil Harvey (philharvey66 at gmail.com)
 
 This library is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
